import math

import numpy as np
import termplotlib as tpl


def histogram(results):
    res = []
    for result in results:
        for image in result:
            if type(image[2]) is not str:
                res.append(abs(float(image[2])))
    if not res:
        print("\nNo angles found, so no histogram.")
        return
    counts, bin_edges = np.histogram(res, bins=20, range=(0, int(math.ceil(max(res)))))

    labels = [
        "{:.2f}° - {:.2f}°".format(bin_edges[k], bin_edges[k + 1])
        for k in range(len(bin_edges) - 1)
    ]
    fig = tpl.figure()
    fig.barh(
        counts,
        labels=labels,
        max_width=40,
        bar_width=1,
        show_vals=True,
        force_ascii=False,
    )

    print("\n=== Skew statistics ===")
    fig.show()

    print(f"Samples: {len(res)}")
    print(f"50th percentile: {np.percentile(res, 50):1.2f}°")
    print(f"90th percentile: {np.percentile(res, 90):1.2f}°")
    print(f"99th percentile: {np.percentile(res, 99):1.2f}°")



import glob
import os
from shutil import rmtree

import pytest


@pytest.fixture(scope="session")
def clean_sampledir():
    os.chdir("samples")
    filelist = glob.glob(".coverage.*")
    for f in filelist:
        os.remove(f)
    os.chdir("..")


@pytest.fixture
def sampledir():
    old_cwd = os.getcwd()
    os.chdir("samples")
    if os.path.isdir("out"):
        rmtree("out")
    if os.path.isdir("debug"):
        rmtree("debug")
    yield
    if os.path.isdir("out"):
        rmtree("out")
    if os.path.isdir("debug"):
        rmtree("debug")
    os.chdir(old_cwd)



import logging
import logging.config
import logging.handlers
import signal
import sys
import traceback
from multiprocessing import Manager, Process


def setup_queue_logging(q):
    h = logging.handlers.QueueHandler(q)
    root = logging.getLogger()
    root.addHandler(h)
    root.setLevel(logging.DEBUG)


def start_logger_process(log_level, results_file):
    logq = Manager().Queue(-1)
    listener = Process(target=_logger_process, args=(logq, log_level, results_file))
    listener.start()
    return logq, listener


def _logger_process(q, log_level, results_file):
    signal.signal(signal.SIGINT, signal.SIG_IGN)
    _setup_logger_process_logging(log_level, results_file)
    try:
        while True:
            record = q.get()
            if record is None:
                break
            logger = logging.getLogger(record.name)
            logger.handle(record)
    except EOFError:
        # print("Logger EOFError, aborting...", file=sys.stderr)
        pass
    except Exception:
        print("Exception in logger process:", file=sys.stderr)
        traceback.print_exc(file=sys.stderr)


def _setup_logger_process_logging(log_level, results_file):
    logd = {
        "version": 1,
        "formatters": {
            "detailed": {
                "class": "logging.Formatter",
                "format": "%(asctime)s %(name)-8s %(levelname)-8s %(processName)-25s %(message)s",
            },
            "raw": {"class": "logging.Formatter", "format": "%(message)s"},
        },
        "handlers": {
            "console": {"class": "logging.StreamHandler", "level": log_level},
            "file": {
                "class": "logging.FileHandler",
                "filename": "hough.log",
                "mode": "a",
                "formatter": "detailed",
                "level": logging.DEBUG,
            },
            "csv": {
                "class": "logging.FileHandler",
                "filename": results_file,
                "mode": "a",
                "formatter": "raw",
            },
        },
        "loggers": {
            "csv": {"handlers": ["csv"]},
            "hough": {"handlers": ["file", "console"]},
        },
        "root": {"level": logging.DEBUG, "handlers": []},
    }
    logging.config.dictConfig(logd)



import logging
import os
from multiprocessing import Pool
from shutil import rmtree

import pytest

from hough import __version__, analyse_file, run
from hough.cli import _abort
from hough.log_utils import start_logger_process


@pytest.mark.usefixtures("clean_sampledir")
def test_version():
    assert __version__


@pytest.mark.usefixtures("sampledir")
def test_histogram_exception(mocker):
    mocker.patch("hough.histogram", new=RuntimeError("Boom"))
    assert run(["--histogram", "av-000.tif"]) == -1


@pytest.mark.usefixtures("sampledir")
def test_null_histogram():
    assert run(["--histogram", "white.jpg"]) == 0


@pytest.mark.usefixtures("sampledir")
def test_abort():
    logq, listener = start_logger_process(logging.DEBUG, "null.csv")
    with Pool(processes=2) as p:
        _abort(p, logq, listener)
    with Pool(processes=2) as p:
        _abort(p, 123, "abc")
    with Pool(processes=2) as p:
        _abort(p, None, None)
    _abort(123, 456, "abc")


@pytest.mark.usefixtures("sampledir")
def test_no_valid_files():
    run(["a", "b", "c"])


@pytest.mark.usefixtures("sampledir")
def test_vert():
    results = analyse_file("av-000.tif")
    res = results[0]
    assert res[0] == "av-000.tif"
    assert res[1] == ""
    assert res[2] > -2 and res[2] <= 0.0
    assert res[3] == ""
    # assert(res[4] == 928)
    # assert(res[5] == 1290)


@pytest.mark.usefixtures("sampledir")
def test_hv_fail():
    run(["av-001.jpg"])


@pytest.mark.usefixtures("sampledir")
def test_multifile_debug():
    rmtree("debug", ignore_errors=True)
    run(["--debug", "av-000.tif", "av-001.jpg"])
    assert os.path.exists("debug")
    assert os.path.exists("hough.log")
    # TODO: assert expected output appears under debug/DATE/whatever


@pytest.mark.usefixtures("sampledir")
def test_unstraightenable():
    run(["--debug", "binder.png"])
    assert os.path.exists("hough.log")
    os.remove("hough.log")
    rmtree("debug", ignore_errors=True)


@pytest.mark.usefixtures("sampledir")
def test_unstraightenable_nodebug():
    run(["binder.png"])


@pytest.mark.usefixtures("sampledir")
def test_histogram_verbose():
    run(["--histogram", "--verbose", "av-000.tif"])


@pytest.mark.usefixtures("sampledir")
def test_rgb_pdf_4_workers():
    if os.path.exists("./__pytest.csv"):
        os.remove("./__pytest.csv")
    run(
        [
            "--debug",
            "--results=./__pytest.csv",
            "--workers=4",
            "Newman_Computer_Exchange_VAX_PC_PDP11_Values.pdf",
        ]
    )
    assert os.path.exists("./__pytest.csv")
    assert os.path.exists("hough.log")
    os.remove("./__pytest.csv")
    os.remove("hough.log")
    rmtree("debug", ignore_errors=True)


@pytest.mark.usefixtures("sampledir")
def test_form_pdf():
    run(["i-9.pdf"])


@pytest.mark.usefixtures("sampledir")
def test_mixed_pdf():
    run(["i-9-paper-version.pdf"])


@pytest.mark.usefixtures("sampledir")
def test_white():
    run(["white.jpg"])


@pytest.mark.usefixtures("sampledir")
def test_unknown():
    run(["../README.md"])


@pytest.mark.usefixtures("sampledir")
def test_broken_out():
    open(".__pytest", "a").close()
    assert run(["--out=.__pytest", "white.jpg"]) == -1
    os.remove(".__pytest")


@pytest.mark.usefixtures("sampledir")
def test_no_args():
    assert run([]) == 0


@pytest.mark.usefixtures("sampledir")
def test_rotate_nothing():
    if os.path.exists(".__pytest"):
        os.remove(".__pytest")
    open(".__pytest", "a").close()
    assert run(["--rotate", "--results=.__pytest"]) == -1
    os.remove(".__pytest")


@pytest.mark.usefixtures("sampledir")
def test_rotate():
    if os.path.exists(".__pytest.dir"):
        rmtree(".__pytest.dir")
    run(
        [
            "--rotate",
            "--out=.__pytest.dir",
            "Newman_Computer_Exchange_VAX_PC_PDP11_Values.pdf",
            "av-000.tif",
        ]
    )
    assert os.path.exists(".__pytest.dir")
    assert os.path.exists(".__pytest.dir/results.csv")
    assert os.path.exists(
        ".__pytest.dir/Newman_Computer_Exchange_VAX_PC_PDP11_Values.pdf"
    )
    assert os.path.exists(".__pytest.dir/av-000.tif")
    rmtree(".__pytest.dir")


@pytest.mark.usefixtures("sampledir")
def test_rotate_from_results():
    if os.path.exists(".__pytest.dir"):
        rmtree(".__pytest.dir")
    assert run(["--rotate", "--out=.__pytest.dir", "--results=newman-results.csv"]) == 0
    assert os.path.exists(
        ".__pytest.dir/Newman_Computer_Exchange_VAX_PC_PDP11_Values.pdf"
    )
    rmtree(".__pytest.dir")



"""
Worker functions for a parallelizable skew analyser.
"""
import logging
import os
import signal

import filetype
import fitz
import numpy as np
from imageio import imread, imwrite
from skimage.color import rgb2gray
from skimage.draw import line_aa
from skimage.exposure import is_low_contrast
from skimage.feature import canny
from skimage.filters import threshold_otsu
from skimage.morphology import binary_dilation
from skimage.transform import probabilistic_hough_line, rescale
from skimage.util import crop, img_as_ubyte, invert

import hough

from . import log_utils


# numpy's little helpers


def grey(x):
    return 0.3 if x else 0.0


def bool_to_255(x):
    return 255 if x else 0


def sum(a):
    return a.sum()


bool_to_255f = np.vectorize(bool_to_255)
greyf = np.vectorize(grey)
hough_prec = np.deg2rad(0.02)
hough_theta_h = np.arange(np.deg2rad(-93.0), np.deg2rad(-87.0), hough_prec)
hough_theta_v = np.arange(np.deg2rad(-3.0), np.deg2rad(3.0), hough_prec)
hough_theta_hv = np.concatenate((hough_theta_v, hough_theta_h))


def hough_angles(pos, neg, orientation="row"):
    height, width = pos.shape
    if orientation == "row":
        axis = 1
        length = int(width * 0.15)
        theta = hough_theta_h
    else:
        axis = 0
        length = int(height * 0.15)
        theta = hough_theta_v
    sums = np.apply_along_axis(sum, axis, neg)
    line = sums.argmax(0)

    # Grab a +/- WINDOW-SIZE strip for evaluation. We've already cropped out the margins.
    if orientation == "row":
        cropped = pos[
            max(line - hough.WINDOW_SIZE, 0) : min(
                line + hough.WINDOW_SIZE, height
            )  # noqa: E203
        ]
    else:
        cropped = pos[
            :,
            max(line - hough.WINDOW_SIZE, 0) : min(
                line + hough.WINDOW_SIZE, width
            ),  # noqa: E203
        ]
    edges = binary_dilation(canny(cropped, 2))
    edges_grey = greyf(edges)

    lines = probabilistic_hough_line(edges, line_length=length, line_gap=2, theta=theta)

    angles = []
    for ((x0, y0), (x1, y1)) in lines:
        # Ensure line is moving rightwards/upwards
        if orientation == "row":
            k = 1 if x1 > x0 else -1
            offset = 0
        else:
            k = 1 if y1 > y0 else -1
            offset = 90
        angles.append(offset - np.rad2deg(np.math.atan2(k * (y1 - y0), k * (x1 - x0))))
        rr, cc, val = line_aa(c0=x0, r0=y0, c1=x1, r1=y1)
        for k, v in enumerate(val):
            edges_grey[rr[k], cc[k]] = (1 - v) * edges_grey[rr[k], cc[k]] + v

    return (angles, sums[line], edges_grey)


def _init_worker(queue, debug_arg, now_arg):
    signal.signal(signal.SIGINT, signal.SIG_IGN)
    log_utils.setup_queue_logging(queue)
    # this is a global only within the multiprocessing Pool workers, not in the main process.
    global debug, now
    debug = debug_arg
    now = now_arg


def get_pages(f):
    """Returns the pages in the file, as a list of tuples of the form:
        [(filename, "mime/type", pagenum), ... ]
    """
    try:
        kind = filetype.guess(f)
    except (FileNotFoundError, IsADirectoryError):
        return []
    if not kind:
        return [(f, None, None)]
    if kind.mime == "application/pdf":
        pdf = fitz.open(f)
        return [(f, kind.mime, x) for x in range(len(pdf))]
    # TODO: Add support for multi-page TIFFs here via
    #   https://imageio.readthedocs.io/en/stable/userapi.html#imageio.mimread
    else:
        return [(f, kind.mime, None)]


def analyse_page(tuple):
    f, mimetype, page = tuple
    logger = logging.getLogger("hough")
    if page is None:
        # not a known multi-image file
        try:
            image = imread(f)
            logger.info(f"Processing {f}...")
            return [analyse_image(f, image, logger)]
        except ValueError as e:
            # Fall through; might be multi-page anyway...
            logger.debug(f"Single-page read of {f} failed: {e}")
    if mimetype == "application/pdf":
        results = []
        doc = fitz.open(f)
        imagelist = doc.getPageImageList(page)
        # TODO: Correctly deal with multiple images on a page (in imagelist)
        for item in imagelist:
            xref = item[0]
            smask = item[1]
            if smask == 0:
                imgdict = doc.extractImage(xref)
                logger.info(f"Processing {f} - page {page+1} - xref {xref}...")
                try:
                    image = imread(imgdict["image"])
                    results.append(analyse_image(f, image, logger, pagenum=(page + 1)))
                except ValueError as e:
                    logger.error(f"Skipping {f} - page {page+1} - xref {xref}: {e}")
            else:
                logger.error(
                    f"Skipping process {f} - page {page+1} - image {xref} (smask=={smask})"
                )
        return results
    else:
        # TODO: support multi-image TIFF with
        #   https://imageio.readthedocs.io/en/stable/userapi.html#imageio.mimread
        logger.error(f"Cannot process {f}: unknown file format")
        return []


def analyse_file(f):
    results = []
    for page in get_pages(f):
        results += analyse_page(page)
    return results


def analyse_image(f, page, logger, pagenum=None):
    global debug, now
    if "debug" not in globals():
        debug = False
    if "now" not in globals():
        import datetime

        now = datetime.datetime.utcnow().strftime("%Y-%m-%dT%H%M%SZ")
    pagenum = int(pagenum) if pagenum is not None else ""

    filename = os.path.basename(f)

    if page.ndim > 2:
        logger.debug(f"{f} p{pagenum} is multichannel - converting to greyscale")
        page = rgb2gray(page)
    pageh, pagew = page.shape
    logger.debug(f"Processing {filename} p{pagenum} ({pagew} x {pageh})...")

    # Remove the margins, which are often dirty due to skewing
    # 0.33" of an 8.5" page is approximately 1/25th
    pos = crop(page, pagew // 25)

    if is_low_contrast(pos):
        logger.debug(f"{filename}  p{pagenum} - low contrast - blank page?")
        return (f, pagenum, "", "", pagew, pageh)

    neg = invert(pos)

    h_angles, v_sums_row, h_edges_grey = hough_angles(pos, neg, "row")
    v_angles, h_sums_col, v_edges_grey = hough_angles(pos, neg, "column")

    angles = []

    if h_angles and v_sums_row > h_sums_col:
        angle = np.median(h_angles)
        if debug:
            imwrite(
                f"debug/{now}/{filename}_{pagenum}_{angle}_hlines.png",
                img_as_ubyte(h_edges_grey),
            )
        logger.debug(f"{filename} p{pagenum} Hough H angle: {angle} deg (median)")
    elif v_angles:
        angle = np.median(v_angles)
        if debug:
            imwrite(
                f"debug/{now}/{filename}_{pagenum}_{angle}_vlines.png",
                img_as_ubyte(v_edges_grey),
            )
        logger.debug(f"{filename} p{pagenum} Hough V angle: {angle} deg (median)")
    else:
        if debug:
            imwrite(
                f"debug/{now}/{filename}_{pagenum}_no_hlines.png",
                img_as_ubyte(h_edges_grey),
            )
            imwrite(
                f"debug/{now}/{filename}_{pagenum}_no_vlines.png",
                img_as_ubyte(v_edges_grey),
            )
        logger.debug(f"{filename} p{pagenum} failed peak sum Hough H/V")

        # We didn't find a good feature at the H or V sum peaks.
        # Let's brutally dilate everything and look for a vertical margin!
        small = rescale(neg, 0.5, anti_aliasing=True)
        t = threshold_otsu(small)
        dilated = binary_dilation(small > t, np.ones((60, 60)))

        edges = canny(dilated, 3)
        edges_grey = greyf(edges)
        lines = probabilistic_hough_line(
            edges, line_length=int(pageh * 0.04), line_gap=6, theta=hough_theta_hv,
        )

        for ((x_0, y_0), (x_1, y_1)) in lines:
            if abs(x_1 - x_0) > abs(y_1 - y_0):
                # angle is <= π/4 from horizontal or vertical
                _, x0, y0, x1, y1 = "H", x_0, y_0, x_1, y_1
            else:
                _, x0, y0, x1, y1 = "V", y_0, -x_0, y_1, -x_1
            # flip angle so that X delta is positive (East quadrants).
            k = 1 if x1 > x0 else -1
            a = np.rad2deg(np.math.atan2(k * (y1 - y0), k * (x1 - x0)))

            # Zero angles are suspicious -- could be a cropping margin.
            # If not, they don't add information anyway.
            if a != 0:
                angles.append(-a)
                rr, cc, val = line_aa(c0=x_0, r0=y_0, c1=x_1, r1=y_1)
                for k, v in enumerate(val):
                    edges_grey[rr[k], cc[k]] = (1 - v) * edges_grey[rr[k], cc[k]] + v

        if angles:
            angle = np.mean(angles)
            if debug:
                imwrite(
                    f"debug/{now}/{filename}_{pagenum}_{angle}_lines_vertical.png",
                    img_as_ubyte(edges_grey),
                )
                imwrite(
                    f"debug/{now}/{filename}_{pagenum}_{angle}_lines_verticaldilated.png",
                    bool_to_255f(img_as_ubyte(dilated)),
                )
            logger.debug(
                f"{filename} p{pagenum} Hough angle V: {angle} deg (mean)  {np.median(angles)} deg (median)"
            )
        else:
            angle = None
            if debug:
                imwrite(
                    f"debug/{now}/{filename}_{pagenum}_dilated.png",
                    bool_to_255f(img_as_ubyte(dilated)),
                )
                imwrite(
                    f"debug/{now}/{filename}_{pagenum}_dilate_edges_grey.png",
                    img_as_ubyte(edges_grey),
                )
                imwrite(
                    f"debug/{now}/{filename}_{pagenum}_dilate_edges.png",
                    bool_to_255f(img_as_ubyte(edges)),
                )
            logger.debug(f"{filename} p{pagenum} failed dilated Hough V")

    return (
        f,
        pagenum,
        angle if angle is not None else "",
        np.var(angles) if angles else "",
        pagew,
        pageh,
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#    This file is part of "hough", which detects skew angles in scanned images
#    Copyright (C) 2016-2020 Toby Thain, toby@telegraphics.com.au
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""hough - straighten scanned pages using the Hough transform.

Usage:
  hough (-h | --help)
  hough [options] [FILE] ...
  hough [options] [--results=<file>] [FILE] ...
  hough (-r | --rotate) [options] [--results=<file>]
  hough (-r | --rotate) [options] [--results=<file>] [FILE] ...

Arguments:
  FILE                          input files to analyze/rotate

Options:
  -h --help                     display this help and exit
  --version                     display the version number and exit
  -v --verbose                  print status messages
  -d --debug                    retain debug image output in debug/ dir
                                (also enables --verbose)
  --histogram                   display rotation angle histogram summary
  -o DIR, --out=DIR             store output results/images in named
                                directory. Directory is created if it
                                does not exist [default: out/TIMESTAMP]
  --results=<file>              save results in FILE under output path,
                                or specify path to results file for
                                rotation [default: results.csv]
  -w <workers> --workers=<#>    specify the number of workers to run
                                simultaneously. Default: total # of CPUs
  -r --rotate                   rotates the files passed on the command
                                line, or if none given, those listed
                                in the results file.
"""

import csv
import datetime
import logging
import os
import sys
import time
from multiprocessing import Pool, cpu_count, freeze_support

from docopt import docopt
from tqdm import tqdm

import hough

from . import log_utils


def _abort(pool=None, log_queue=None, listener=None):
    try:
        if pool:
            pool.close()
            pool.terminate()
            pool.join()
            # this lets the producers drain their log queues
            time.sleep(0.1)
        if log_queue and listener:
            print(
                f"=== Run killed @ {datetime.datetime.utcnow().isoformat()} ===",
                file=sys.stderr,
            )
            try:
                log_queue.put(None)
                listener.join()
            except Exception:
                pass
    except Exception:
        import traceback

        print("Exception during abort:", file=sys.stderr)
        traceback.print_exc(file=sys.stderr)


def _process_args(arguments):
    if arguments.debug:
        arguments["verbose"] = True
        arguments["log_level"] = logging.DEBUG
    elif arguments.verbose:
        arguments["log_level"] = logging.INFO
    else:
        arguments["log_level"] = logging.WARNING
    arguments["pbar_disable"] = arguments.quiet or arguments.verbose or arguments.debug

    arguments["now"] = datetime.datetime.utcnow().strftime("%Y-%m-%dT%H%M%SZ")
    if arguments.out == "out/TIMESTAMP":
        arguments["out"] = f"out/{arguments.now}"
    else:
        arguments["out"] = os.path.abspath(arguments.out)
    if not os.path.isdir(arguments.out):
        try:
            os.makedirs(arguments.out)
        except Exception:
            print(f"Unable to create output directory f{arguments.out}! Exiting...")
            return -1
    if arguments.debug and not os.path.isdir(f"debug/{arguments.now}"):
        os.makedirs(f"debug/{arguments.now}")

    # --rotate needs files to process, or results
    if arguments.rotate and len(arguments.FILE) == 0:
        if (
            not os.path.exists(arguments.results)
            or os.path.getsize(arguments.results) == 0
        ):
            print(
                f"No results in {arguments.results} to rotate and no files specified!"
            )
            return -1
    # --rotate with args, or files given. Was a results file path specified?
    elif os.path.split(arguments.results)[0]:
        arguments["results"] = os.path.abspath(arguments.results)
    else:
        arguments["results"] = os.path.join(arguments.out, arguments.results)

    if arguments.workers:
        arguments["workers"] = int(arguments.workers)
    else:
        arguments["workers"] = cpu_count()

    return arguments


def run(argv=sys.argv[1:]):
    if len(argv) == 0:
        print(__doc__.strip("\n"))
        return 0
    arguments = docopt(__doc__, argv=argv, version=hough.__version__, more_magic=True)
    arguments = _process_args(arguments)
    if arguments == -1:
        return -1

    logq, listener = log_utils.start_logger_process(
        arguments.log_level, arguments.results
    )

    results = []

    # The pool that launched 1,000 Houghs...
    # We do it this way, not with a map, until https://github.com/tqdm/tqdm/issues/548 is fixed
    with Pool(
        processes=arguments.workers,
        initializer=hough.analyse._init_worker,
        initargs=(logq, arguments.debug, arguments.now,),
    ) as p:
        try:
            log_utils.setup_queue_logging(logq)
            logger = logging.getLogger("hough")
            logger.info(
                f"=== Run started @ {datetime.datetime.utcnow().isoformat()} ==="
            )

            pages = []
            for f in arguments.FILE:
                pages += hough.get_pages(f)

            if pages:
                with tqdm(
                    total=len(pages),
                    disable=arguments.pbar_disable,
                    unit="pg",
                    desc="Analysis: ",
                ) as pbar:
                    for i, result in enumerate(
                        p.imap_unordered(hough.analyse_page, pages)
                    ):
                        pbar.update()
                        results.append(result)
            p.close()
            p.join()
        except KeyboardInterrupt:
            import sys

            print("Caught KeyboardInterrupt, terminating workers...", file=sys.stderr)
            _abort(p, logq, listener)
            return -1

    logger_csv = logging.getLogger("csv")
    if not os.path.exists(arguments.results) or os.path.getsize(arguments.results) == 0:
        logger_csv.info(
            '"Input File","Page Number","Computed angle","Variance of computed angles","Image width (px)","Image height (px)"'
        )

    read_csv = False
    # might be rotating a previously generated csv file...
    if (
        not results
        and os.path.exists(arguments.results)
        and os.path.getsize(arguments.results) > 0
    ):
        with open(arguments.results, newline="") as csvfile:
            reader = csv.reader(csvfile)
            for row in reader:
                if row[0] == "Input File":
                    continue
                for idx in [1, 4, 5]:
                    row[idx] = int(row[idx]) if row[idx] else ""
                for idx in [2, 3]:
                    row[idx] = float(row[idx]) if row[idx] else ""
                results.append([tuple(row)])
        read_csv = True

    if arguments.histogram:
        try:
            hough.histogram(results)
        except Exception:
            import sys
            import traceback

            logger.error(f"Exception in histogram process: \n{traceback.format_exc()}")
            _abort(None, logq, listener)
            return -1

    dictresults = {}
    num_pages = 0
    for result in results:
        for image in result:
            num_pages += 1
            dictresults.setdefault(image[0], []).append(image)
            (fname, pagenum, angle, variance, pagew, pageh) = image
            if not read_csv:
                logger_csv.info(
                    '"{}",{},{},{},{},{}'.format(
                        fname, pagenum, angle, variance, pagew, pageh,
                    )
                )

    if arguments.rotate:
        with tqdm(
            total=num_pages,
            disable=arguments.pbar_disable,
            unit="pg",
            desc="Rotation: ",
        ) as pbar:
            for f in dictresults:
                for i, result in enumerate(
                    hough.rotate(
                        sorted(dictresults[f], key=lambda x: int(x[1]) if x[1] else 0),
                        arguments.out,
                        generator=True,
                    )
                ):
                    pbar.update()

    logger.info(f"=== Run ended @ {datetime.datetime.utcnow().isoformat()} ===")

    # end logging process
    logq.put(None)
    listener.join()

    return 0


if __name__ == "__main__":
    freeze_support()
    exit(run())



from .analyse import analyse_file, analyse_page, get_pages
from .cli import run
from .rotate import rotate
from .stats import histogram


try:
    from importlib.metadata import version, PackageNotFoundError  # type: ignore
except ImportError:  # pragma: no cover
    from importlib_metadata import version, PackageNotFoundError  # type: ignore


try:
    __version__ = version(__name__)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"

WINDOW_SIZE = 150

__all__ = ["analyse_file", "analyse_page", "get_pages", "run", "rotate", "histogram"]



"""
Worker functions for a parallelizable deskewer.
"""
import logging
import os

import filetype
import fitz
from imageio import imread, imwrite
from scipy.ndimage import interpolation


def rotate(imagelist, out, generator=False):
    """Actually rotates a single file of 1+ images.
    Rotated file has the same name and is placed under the {out} subdirectory
    of the cwd.
    """
    logger = logging.getLogger("hough")
    guess = filetype.guess(imagelist[0][0])
    if guess and guess.mime == "application/pdf":
        newdoc = fitz.open()
    else:
        newdoc = None
    filename = imagelist[0][0]
    filen, ext = os.path.splitext(os.path.basename(filename))
    kind = filetype.guess(filename)
    for image in imagelist:
        page = int(image[1]) if image[1] else ""
        angle = float(image[2]) if image[2] else 0.0
        if not page:
            # single-image file, not a container
            logger.info(f"Rotating {filename}...")
            img = imread(image[0])
            fixed = interpolation.rotate(img, -angle, mode="nearest", reshape=False)
            imwrite(f"{out}/{filen}{ext}", fixed)
        else:
            if kind and kind.mime == "application/pdf":
                doc = fitz.open(image[0])
                imagelist = doc.getPageImageList(page - 1)
                # TODO: Correctly deal with multiple images on a page
                for item in imagelist:
                    xref = item[0]
                    smask = item[1]
                    if smask == 0:
                        imgdict = doc.extractImage(xref)
                        logger.info(
                            f"Rotating {filename} - page {page} - xref {xref}..."
                        )
                        try:
                            img = imread(imgdict["image"])
                            imgext = imgdict["ext"]
                            fixed = interpolation.rotate(
                                img, -angle, mode="nearest", reshape=False
                            )
                            imgbytes = imwrite("<bytes>", fixed, format=imgext)
                            imgdoc = fitz.open(stream=imgbytes, filetype=imgext)
                            rect = imgdoc[0].rect
                            pdfbytes = imgdoc.convertToPDF()
                            imgdoc.close()
                            imgPDF = fitz.open("pdf", pdfbytes)
                            page = newdoc.newPage(width=rect.width, height=rect.height)
                            page.showPDFpage(rect, imgPDF, 0)
                        except ValueError as e:
                            logger.error(
                                f"Skipping rotating {filename} - page {page} - xref {xref}: {e}"
                            )
                    else:
                        logger.error(
                            f"Skipping process {filename} - page {page} - image {xref} (smask=={smask})"
                        )
            # TODO: deal with other multi-image formats
            else:
                logger.error(
                    f"Skipping file {filename} - unknown multi-page file format"
                )
        yield 1
    if newdoc:
        newdoc.save(f"{out}/{filen}{ext}")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['hough', 'tests']

package_data = \
{'': ['*']}

install_requires = \
['coverage[toml]>=5.0.4,<6.0.0',
 'docopt-ng>=0.7.2,<0.8.0',
 'filetype>=1.0.6,<2.0.0',
 'imageio>=2.8.0,<3.0.0',
 'numpy>=1.18.2,<2.0.0',
 'pymupdf>=1.17.0,<2.0.0',
 'scikit-image>=0.16.2,<0.18.0',
 'scipy>=1.4.1,<2.0.0',
 'termplotlib>=0.3.0,<0.4.0',
 'tqdm>=4.43.0,<5.0.0']

extras_require = \
{':python_version < "3.8"': ['importlib_metadata>=2.0.0,<3.0.0']}

entry_points = \
{'console_scripts': ['hough = hough.cli:run']}

setup_kwargs = {
    'name': 'hough',
    'version': '0.2.6',
    'description': 'Skew detection and correction in scanned images',
    'long_description': '# hough - Skew detection in scanned images\n\n<p align="center">\n<a href="https://github.com/wohali/hough/actions"><img alt="Actions Status" src="https://github.com/wohali/hough/workflows/Tests/badge.svg"></a>\n<a href="https://pypi.org/project/hough/"><img alt="PyPI" src="https://img.shields.io/pypi/v/hough"></a>\n<a href="https://pypi.org/project/hough/"><img alt="PyPI - Python Version" src="https://img.shields.io/pypi/pyversions/hough"></a>\n<a href="https://github.com/wohali/hough/blob/master/COPYING"><img src="https://img.shields.io/github/license/wohali/hough.svg" alt="GPL v2.0 License" /></a>\n<a href="https://github.com/psf/black"><img alt="Code style: black" src="https://img.shields.io/badge/code%20style-black-000000.svg"></a>\n<a href="https://codecov.io/gh/wohali/hough"><img alt="Coverage stats" src="https://codecov.io/gh/wohali/hough/branch/master/graph/badge.svg" /></a>\n</p>\n\n_Hough_ finds skew angles in scanned document pages, using the Hough transform.\n\nIt is oriented to batch processing, and can make use of multiple cores. (You\'ll\nwant this - analysis and image processing is very CPU intensive!)\n\n# Installation and usage\n\n## Installation\n\n```\npip install -U pip\npip install hough\n```\n\nThe first line is required to update `pip` to a new enough version to be\ncompatible with `manylinux` wheel packaging, required for PyMuPDF.\n\nOlder versions of `pip` are fine, but you\'ll have to install MuPDF, its\nheaders, and a compiler first, so PyMuPDF can be compiled locally.\n\n## Usage\n\nTo get started right away, here\'s some examples.\n\nGenerate angles (in CSV form) for a bunch of TIFF page images, one page per file:\n\n```\nhough --csv in/*.tif\n```\n\nThe same, but for a PDF file, and display a histogram at the end:\n\n```\nhough --histogram Able_Attach_Sep83.pdf\n```\n\nThe same, but show progress while running:\n\n```\nhough -v --histogram Able_Attach_Sep83.pdf\n```\n\n\nThe deskewing results are placed in the `results.csv` file. Example:\n\n```csv\n"Input File","Page Number","Computed angle","Variance of computed angles","Image width (px)","Image height (px)"\n"/home/toby/my-pages/orig/a--0000.pgm.tif",,-0.07699791151672428,0.001073874144832815,5014,6659\n"/home/toby/my-pages/orig/a--0001.pgm.tif",,,,5018,6630\n"/home/toby/my-pages/orig/a--0002.pgm.tif",,0.24936351676615068,0.005137031681286154,5021,6629\n"/home/toby/my-pages/orig/a--0003.pgm.tif",,,,5020,6608\n"/home/toby/my-pages/orig/a--0004.pgm.tif",,-0.037485115754500545,0.025945115897015238,5021,6616\n```\n\nThe program should work on various image input formats, and with both grey scale\nand RGB images. _Hough_ works best with images ≥300dpi.\n\nHere\'s a histogram sample:\n\n```\n=== Skew statistics ===\n0.00° - 0.10°  [57]  ████████████████████████████████████████\n0.10° - 0.20°  [39]  ███████████████████████████▍\n0.20° - 0.30°  [30]  █████████████████████\n0.30° - 0.40°  [30]  █████████████████████\n0.40° - 0.50°  [11]  ███████▊\n0.50° - 0.60°  [11]  ███████▊\n0.60° - 0.70°  [ 3]  ██▏\n0.70° - 0.80°  [ 4]  ██▊\n0.80° - 0.90°  [ 0]\n0.90° - 1.00°  [ 1]  ▊\n1.00° - 1.10°  [ 1]  ▊\n1.10° - 1.20°  [ 0]\n1.20° - 1.30°  [ 1]  ▊\n1.30° - 1.40°  [ 1]  ▊\n1.40° - 1.50°  [ 1]  ▊\n1.50° - 1.60°  [ 2]  █▍\n1.60° - 1.70°  [ 0]\n1.70° - 1.80°  [ 1]  ▊\n1.80° - 1.90°  [ 2]  █▍\n1.90° - 2.00°  [ 0]\nSamples: 195\n50th percentile: 0.20°\n90th percentile: 0.55°\n99th percentile: 1.77°\n```\n\n## Command line options\n\nYou can list them by running `hough --help`:\n\n```\nhough - straighten scanned pages using the Hough transform.\n\nUsage:\n  hough (-h | --help)\n  hough [options] [FILE] ...\n  hough [options] [--results=<file>] [FILE] ...\n  hough (-r | --rotate) [options] [--results=<file>]\n  hough (-r | --rotate) [options] [--results=<file>] [FILE] ...\n\nArguments:\n  FILE                          input files to analyze/rotate\n\nOptions:\n  -h --help                     display this help and exit\n  --version                     display the version number and exit\n  -v --verbose                  print status messages\n  -d --debug                    retain debug image output in debug/ dir\n                                (also enables --verbose)\n  --histogram                   display rotation angle histogram summary\n  -o DIR, --out=DIR             store output results/images in named\n                                directory. Directory is created if it\n                                does not exist [default: out/TIMESTAMP]\n  --results=<file>              save results in FILE under output path,\n                                or specify path to results file for\n                                rotation [default: results.csv]\n  -w <workers> --workers=<#>    specify the number of workers to run\n                                simultaneously. Default: total # of CPUs\n  -r --rotate                   rotates the files passed on the command\n                                line, or if none given, those listed\n                                in the results file.\n```\n\n# Examples\n\nJust about all of [these files](http://docs.telegraphics.com.au/) have been\ndeskewed this way.\n\n# Getting the best results\n\n### NOTE: This is a beta product!\n\nThere\'s a few guidelines you should follow to get the best deskewing results\nfrom your document scans:\n\n1. Bilevel (black-and-white) bitmaps will produce lower quality results.\n   For best results, scan to greyscale or RGB first, deskew with _Hough_, then\n   reduce the colour depth to bilevel.\n1. Hough deskewing is an inexact process, with many heuristics discovered\n   by trial and error. _Hough_ may not work well on your material without tuning\n   and further modification. (We\'d love your pull requests!)\n\n## Debugging output\n\nYou can spy on _Hough_\'s attempts to perform deskewing by passing the `--debug`\nflag on the command line. The generated images, and any detected lines in them,\nare placed in the `debug/<datetime>/` directory.\n\nNote that _Hough_ cannot always determine a skew for a page (e.g. blank pages\nin particular), and will very occasionally get the skew wrong (depending on\nsource material). It\'s worth reviewing these images if _Hough_ makes a bad\ndecision on your scans. Please submit these files along with the original image\nwhen filing an issue!\n\n## Recommended scanners\n\nThe authors have tested this software with output from the following scanners:\n\n* Fujitsu fi-4530C, USB\n  * Fast\n  * Cheap on eBay\n  * Requires a Windows XP VirtualBox for drivers\n* Brother ADS-2700W, USB + Ethernet + WiFi\n  * Fast\n  * Can scan directly to the network or to a memory stick\n  * Factory reconditioned models stilll available (March 2020)\n  * Very low skew out of the box\n* Epson WF-7610, USB + Ethernet + WiFi\n  * 11"x17" and duplex capable\n  * Can scan directly to the network or to a memory stick\n\n# Developing\n\nFirst, clone this repo.\n\nYou\'ll need to install [Poetry](https://python-poetry.org/docs/#installation),\nthen run:\n\n```\npoetry run pip install -U pip setuptools\npoetry install\npoetry shell\n```\n\n# License notice\n\n```\nThis file is part of "hough", which detects skew angles in scanned images\nCopyright (C) 2016-2020 Toby Thain <toby@telegraphics.com.au>\nCopyright (C) 2020 Joan Touzet <wohali@apache.org>\n\nThis program is free software; you can redistribute it and/or modify\nit under the terms of the GNU General Public License as published by\nthe Free Software Foundation; either version 2 of the License, or\n(at your option) any later version.\n\nThis program is distributed in the hope that it will be useful,\nbut WITHOUT ANY WARRANTY; without even the implied warranty of\nMERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\nGNU General Public License for more details.\n\nYou should have received a copy of the GNU General Public License\nalong with this program; if not, write to the Free Software\nFoundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA\n```\n',
    'author': 'qu1j0t3',
    'author_email': 'support@telegraphics.com.au',
    'maintainer': 'Joan Touzet',
    'maintainer_email': 'wohali@apache.org',
    'url': 'https://github.com/wohali/hough',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)
